/* -------------- declaration section -------------- */

%{

#ifdef	WIN32
#pragma warning(disable : 4786)
#include <cstdio>
#include <memory>
namespace std
{ using ::size_t; 
  using ::malloc;
  using ::free;
}
#endif

#include "librib.h"
using namespace librib;

#include "libribtypes.h"

#include <algorithm>
#include <iostream>
#include <map>
#include <string>
#include <vector>
#include <cassert>

namespace librib
{

void ParserDeclare(librib::RendermanInterface& CallbackInterface, const std::string Name, const std::string Type);
extern std::string ParseStreamName;
extern RendermanInterface* ParseCallbackInterface;
extern std::ostream* ParseErrorStream;
extern unsigned int ParseLineNumber;
extern bool ParseSucceeded;
extern bool fRequest;
extern bool fParams;

}; // namespace librib

%}

%union
{
	int itype;
	float ftype;
	char* stype;
	char ctype;

	IntegerArray* aitype;
	FloatArray* aftype;
	StringArray* astype;
	TokenValuePair* tvtype;
	TokenValuePairs* atvtype;
}

%{

extern int yylex();

static void yyerror(const std::string Message);

static void ExpectRequest();
static void ExpectParams();
static void DiscardStringValue(char* const Value);
static void DiscardTokenValuePairs(TokenValuePairs* const Value);
static void DiscardArrayValue(Array* const Value);

static void DeclareLightHandle(const RendermanInterface::RtInt ID, const RendermanInterface::RtLightHandle Handle);
static RendermanInterface::RtLightHandle LookupLightHandle(const RendermanInterface::RtInt ID);

static ParameterType DeclareParameterType(const std::string& Name, const std::string& Type);
static ParameterType LookupParameterType(const std::string& Name);

static RendermanInterface::RtInt ColorSamples = 3;

%}

%token <ctype> EOL_TOKEN EOF_TOKEN CHAR_TOKEN UNKNOWN_TOKEN
%token <ftype> FLOAT_TOKEN
%token <itype> INTEGER_TOKEN
%token <stype> STRING_TOKEN
%type <itype> integer
%type <ftype> float
%type <stype> string
%type <aftype> scalar_array scalar_array_or_empty
%type <aftype> floats float_array
%type <aitype> integers integer_array integer_array_or_empty
%type <astype> strings string_array string_array_or_empty 
%type <tvtype> tvpair
%type <atvtype> tvpairs
%type <atvtype> opttvpairs

%token	<ctype>	REQUEST_TOKEN_AREALIGHTSOURCE
%token	<ctype>	REQUEST_TOKEN_ATMOSPHERE
%token	<ctype>	REQUEST_TOKEN_ATTRIBUTE
%token	<ctype>	REQUEST_TOKEN_ATTRIBUTEBEGIN
%token	<ctype>	REQUEST_TOKEN_ATTRIBUTEEND
%token	<ctype>	REQUEST_TOKEN_BASIS
%token	<ctype>	REQUEST_TOKEN_BESSELFILTER
%token	<ctype>	REQUEST_TOKEN_BOUND
%token	<ctype>	REQUEST_TOKEN_BOXFILTER
%token	<ctype>	REQUEST_TOKEN_CATMULLROMFILTER
%token	<ctype>	REQUEST_TOKEN_CLIPPING
%token	<ctype>	REQUEST_TOKEN_COLOR
%token	<ctype>	REQUEST_TOKEN_COLORSAMPLES
%token	<ctype>	REQUEST_TOKEN_CONCATTRANSFORM
%token	<ctype>	REQUEST_TOKEN_CONE
%token	<ctype>	REQUEST_TOKEN_COORDINATESYSTEM
%token	<ctype>	REQUEST_TOKEN_COORDSYSTRANSFORM
%token	<ctype>	REQUEST_TOKEN_CROPWINDOW
%token	<ctype>	REQUEST_TOKEN_CYLINDER
%token	<ctype>	REQUEST_TOKEN_DECLARE
%token	<ctype>	REQUEST_TOKEN_DEFORMATION
%token	<ctype>	REQUEST_TOKEN_DEPTHOFFIELD
%token	<ctype>	REQUEST_TOKEN_DETAIL
%token	<ctype>	REQUEST_TOKEN_DETAILRANGE
%token	<ctype>	REQUEST_TOKEN_DISK
%token	<ctype>	REQUEST_TOKEN_DISKFILTER
%token	<ctype>	REQUEST_TOKEN_DISPLACEMENT
%token	<ctype>	REQUEST_TOKEN_DISPLAY
%token	<ctype>	REQUEST_TOKEN_ERRORABORT
%token	<ctype>	REQUEST_TOKEN_ERRORHANDLER
%token	<ctype>	REQUEST_TOKEN_ERRORIGNORE
%token	<ctype>	REQUEST_TOKEN_ERRORPRINT
%token	<ctype>	REQUEST_TOKEN_EXPOSURE
%token	<ctype>	REQUEST_TOKEN_EXTERIOR
%token	<ctype>	REQUEST_TOKEN_FORMAT
%token	<ctype>	REQUEST_TOKEN_FRAMEASPECTRATIO
%token	<ctype>	REQUEST_TOKEN_FRAMEBEGIN
%token	<ctype>	REQUEST_TOKEN_FRAMEEND
%token	<ctype>	REQUEST_TOKEN_GAUSSIANFILTER
%token	<ctype>	REQUEST_TOKEN_GENERALPOLYGON
%token	<ctype>	REQUEST_TOKEN_GEOMETRICAPPROXIMATION
%token	<ctype>	REQUEST_TOKEN_GEOMETRY
%token	<ctype>	REQUEST_TOKEN_HIDER
%token	<ctype>	REQUEST_TOKEN_HYPERBOLOID
%token	<ctype>	REQUEST_TOKEN_IDENTITY
%token	<ctype>	REQUEST_TOKEN_ILLUMINATE
%token	<ctype>	REQUEST_TOKEN_IMAGER
%token	<ctype>	REQUEST_TOKEN_INTERIOR
%token	<ctype>	REQUEST_TOKEN_LIGHTSOURCE
%token	<ctype>	REQUEST_TOKEN_MAKEBUMP
%token	<ctype>	REQUEST_TOKEN_MAKECUBEFACEENVIRONMENT
%token	<ctype>	REQUEST_TOKEN_MAKELATLONGENVIRONMENT
%token	<ctype>	REQUEST_TOKEN_MAKESHADOW
%token	<ctype>	REQUEST_TOKEN_MAKETEXTURE
%token	<ctype>	REQUEST_TOKEN_MATTE
%token	<ctype>	REQUEST_TOKEN_MOTIONBEGIN
%token	<ctype>	REQUEST_TOKEN_MOTIONEND
%token	<ctype>	REQUEST_TOKEN_NUPATCH
%token	<ctype>	REQUEST_TOKEN_OBJECTBEGIN
%token	<ctype>	REQUEST_TOKEN_OBJECTEND
%token	<ctype>	REQUEST_TOKEN_OBJECTINSTANCE
%token	<ctype>	REQUEST_TOKEN_OPACITY
%token	<ctype>	REQUEST_TOKEN_OPTION
%token	<ctype>	REQUEST_TOKEN_ORIENTATION
%token	<ctype>	REQUEST_TOKEN_PARABOLOID
%token	<ctype>	REQUEST_TOKEN_PATCH
%token	<ctype>	REQUEST_TOKEN_PATCHMESH
%token	<ctype>	REQUEST_TOKEN_PERSPECTIVE
%token	<ctype>	REQUEST_TOKEN_PIXELFILTER
%token	<ctype>	REQUEST_TOKEN_PIXELSAMPLES
%token	<ctype>	REQUEST_TOKEN_PIXELVARIANCE
%token	<ctype>	REQUEST_TOKEN_POINTSGENERALPOLYGONS
%token	<ctype>	REQUEST_TOKEN_POINTSPOLYGONS
%token	<ctype>	REQUEST_TOKEN_POLYGON
%token	<ctype>	REQUEST_TOKEN_PROCEDURAL
%token	<ctype>	REQUEST_TOKEN_PROJECTION
%token	<ctype>	REQUEST_TOKEN_QUANTIZE
%token	<ctype>	REQUEST_TOKEN_RELATIVEDETAIL
%token	<ctype>	REQUEST_TOKEN_REVERSEORIENTATION
%token	<ctype>	REQUEST_TOKEN_ROTATE
%token	<ctype>	REQUEST_TOKEN_SCALE
%token	<ctype>	REQUEST_TOKEN_SCREENWINDOW
%token	<ctype>	REQUEST_TOKEN_SHADINGINTERPOLATION
%token	<ctype>	REQUEST_TOKEN_SHADINGRATE
%token	<ctype>	REQUEST_TOKEN_SHUTTER
%token	<ctype>	REQUEST_TOKEN_SIDES
%token	<ctype>	REQUEST_TOKEN_SINCFILTER
%token	<ctype>	REQUEST_TOKEN_SKEW
%token	<ctype>	REQUEST_TOKEN_SOLIDBEGIN
%token	<ctype>	REQUEST_TOKEN_SOLIDEND
%token	<ctype>	REQUEST_TOKEN_SPHERE
%token	<ctype>	REQUEST_TOKEN_SUBDIVISIONMESH
%token	<ctype>	REQUEST_TOKEN_SURFACE
%token	<ctype>	REQUEST_TOKEN_TEXTURECOORDINATES
%token	<ctype>	REQUEST_TOKEN_TORUS
%token	<ctype>	REQUEST_TOKEN_TRANSFORM
%token	<ctype>	REQUEST_TOKEN_TRANSFORMBEGIN
%token	<ctype>	REQUEST_TOKEN_TRANSFORMEND
%token	<ctype>	REQUEST_TOKEN_TRANSFORMPOINTS
%token	<ctype>	REQUEST_TOKEN_TRANSLATE
%token	<ctype>	REQUEST_TOKEN_TRIANGLEFILTER
%token	<ctype>	REQUEST_TOKEN_TRIMCURVE
%token	<ctype>	REQUEST_TOKEN_VERSION
%token	<ctype>	REQUEST_TOKEN_WORLDBEGIN
%token	<ctype>	REQUEST_TOKEN_WORLDEND

%type	<ctype>	arealightsource
%type	<ctype>	atmosphere
%type	<ctype>	attribute
%type	<ctype>	attributebegin
%type	<ctype>	attributeend
%type	<ctype>	basis
%type	<ctype>	bound
%type	<ctype>	clipping
%type	<ctype>	color
%type	<ctype>	colorsamples
%type	<ctype>	concattransform
%type	<ctype>	cone
%type	<ctype>	coordinatesystem
%type	<ctype>	coordsystransform
%type	<ctype>	cropwindow
%type	<ctype>	cylinder
%type	<ctype>	declare
%type	<ctype>	deformation
%type	<ctype>	depthoffield
%type	<ctype>	detail
%type	<ctype>	detailrange
%type	<ctype>	disk
%type	<ctype>	displacement
%type	<ctype>	display
%type	<ctype>	errorabort
%type	<ctype>	errorhandler
%type	<ctype>	errorignore
%type	<ctype>	errorprint
%type	<ctype>	exposure
%type	<ctype>	exterior
%type	<ctype>	format
%type	<ctype>	frameaspectratio
%type	<ctype>	framebegin
%type	<ctype>	frameend
%type	<ctype>	generalpolygon
%type	<ctype>	geometricapproximation
%type	<ctype>	geometry
%type	<ctype>	hider
%type	<ctype>	hyperboloid
%type	<ctype>	identity
%type	<ctype>	illuminate
%type	<ctype>	imager
%type	<ctype>	interior
%type	<ctype>	lightsource
%type	<ctype>	makebump
%type	<ctype>	makecubefaceenvironment
%type	<ctype>	makelatlongenvironment
%type	<ctype>	makeshadow
%type	<ctype>	maketexture
%type	<ctype>	matte
%type	<ctype>	motionbegin
%type	<ctype>	motionend
%type	<ctype>	nupatch
%type	<ctype>	objectbegin
%type	<ctype>	objectend
%type	<ctype>	objectinstance
%type	<ctype>	opacity
%type	<ctype>	option
%type	<ctype>	orientation
%type	<ctype>	paraboloid
%type	<ctype>	patch
%type	<ctype>	patchmesh
%type	<ctype>	perspective
%type	<ctype>	pixelfilter
%type	<ctype>	pixelsamples
%type	<ctype>	pixelvariance
%type	<ctype>	pointsgeneralpolygons
%type	<ctype>	pointspolygons
%type	<ctype>	polygon
%type	<ctype>	projection
%type	<ctype>	quantize
%type	<ctype>	relativedetail
%type	<ctype>	reverseorientation
%type	<ctype>	rotate
%type	<ctype>	scale
%type	<ctype>	screenwindow
%type	<ctype>	shadinginterpolation
%type	<ctype>	shadingrate
%type	<ctype>	shutter
%type	<ctype>	sides
%type	<ctype>	skew
%type	<ctype>	solidbegin
%type	<ctype>	solidend
%type	<ctype>	sphere
%type	<ctype>	subdivisionmesh
%type	<ctype>	surface
%type	<ctype>	texturecoordinates
%type	<ctype>	torus
%type	<ctype>	transform
%type	<ctype>	transformbegin
%type	<ctype>	transformend
%type	<ctype>	transformpoints
%type	<ctype>	translate
%type	<ctype>	trimcurve
%type	<ctype>	version
%type	<ctype>	worldbegin
%type	<ctype>	worldend

/* -------------- rules section -------------- */
%%
file : requests
			{ YYABORT; }
	;

requests : request
	|	requests request
	;

request : complete_request
			{ ExpectRequest(); }

complete_request
	:	version float
			{  }
	|	declare string string
			{
				ParserDeclare(*ParseCallbackInterface, $2, $3);
				DiscardStringValue($2);
				DiscardStringValue($3);
			}
	|	framebegin integer
			{ ParseCallbackInterface->RiFrameBegin($2); }
	|	frameend
			{ ParseCallbackInterface->RiFrameEnd(); }
	|	worldbegin
			{ ParseCallbackInterface->RiWorldBegin(); }
	|	worldend
			{ ParseCallbackInterface->RiWorldEnd(); /* YYACCEPT; */ }
	|	format integer integer float
			{ ParseCallbackInterface->RiFormat($2, $3, $4); }
	|	frameaspectratio float
			{ ParseCallbackInterface->RiFrameAspectRatio($2); }
	|	screenwindow float float float float
			{ ParseCallbackInterface->RiScreenWindow($2, $3, $4, $5); }
	|	cropwindow float float float float
			{ ParseCallbackInterface->RiCropWindow($2, $3, $4, $5); }
	|	projection string opttvpairs
			{
				ParseCallbackInterface->RiProjectionV($2, $3->Count(), $3->Tokens(), $3->Values());
				DiscardStringValue($2);
				DiscardTokenValuePairs($3);
			}
	|	clipping float float
			{ ParseCallbackInterface->RiClipping($2, $3); }
	|	depthoffield float float float
			{ ParseCallbackInterface->RiDepthOfField($2, $3, $4); }
	|	shutter float float
			{ ParseCallbackInterface->RiShutter($2, $3); }
	|	pixelvariance float
			{ ParseCallbackInterface->RiPixelVariance($2); }
	|	pixelsamples float float
			{ ParseCallbackInterface->RiPixelSamples($2, $3); }
	|	pixelfilter string float float
			{
				RendermanInterface::RtFilterFunc pFilter=RiGaussianFilter;
				// TODO: names shoudl be stored elsewhere.
				if(strcmp($2,"box")==0)		pFilter=RiBoxFilter;
				if(strcmp($2,"triangle")==0)	pFilter=RiTriangleFilter;
				if(strcmp($2,"sinc")==0)		pFilter=RiSincFilter;
				if(strcmp($2,"catmull-rom")==0) pFilter=RiCatmullRomFilter;
				if(strcmp($2,"disk")==0)		pFilter=RiDiskFilter;
				if(strcmp($2,"bessel")==0)		pFilter=RiBesselFilter;
				ParseCallbackInterface->RiPixelFilter(pFilter, $3, $4);

				DiscardStringValue($2);
			}
	|	exposure float float
			{ ParseCallbackInterface->RiExposure($2, $3); }
	|	imager string opttvpairs
			{
				ParseCallbackInterface->RiImagerV($2, $3->Count(), $3->Tokens(), $3->Values());
				DiscardStringValue($2);
				DiscardTokenValuePairs($3);
			}
	|	quantize string integer integer integer float
			{
				ParseCallbackInterface->RiQuantize((char*)($2), $3, $4, $5, $6);
				DiscardStringValue($2);
			}
	|	display string string string opttvpairs
			{
				ParseCallbackInterface->RiDisplayV($2, const_cast<char*>($3), const_cast<char*>($4), $5->Count(), $5->Tokens(), $5->Values());
				DiscardStringValue($2);
				DiscardStringValue($3);
				DiscardStringValue($4);
				DiscardTokenValuePairs($5);
			}
	|	hider string opttvpairs
			{
				ParseCallbackInterface->RiHiderV(const_cast<char*>($2), $3->Count(), $3->Tokens(), $3->Values());
				DiscardStringValue($2);
				DiscardTokenValuePairs($3);
			}
	|	colorsamples scalar_array scalar_array
			{
				ColorSamples = $2->Count();
				ParseCallbackInterface->RiColorSamples($2->Count(), &(*$2)[0], &(*$3)[0]);
				DiscardArrayValue($2);
				DiscardArrayValue($3);
			}
	|	relativedetail float
			{ ParseCallbackInterface->RiRelativeDetail($2); }
	|	option string opttvpairs
			{
				ParseCallbackInterface->RiOptionV($2, $3->Count(), $3->Tokens(), $3->Values());
				DiscardStringValue($2);
				DiscardTokenValuePairs($3);
			}
	|	attributebegin
			{ ParseCallbackInterface->RiAttributeBegin(); }
	|	attributeend
			{ ParseCallbackInterface->RiAttributeEnd(); }
	|	color scalar_array
			{
				if($2->Count() == ColorSamples)
					ParseCallbackInterface->RiColor(&(*$2)[0]);
				else
					yyerror("Incorrect Number of Color Samples");
					
				DiscardArrayValue($2);
			}
	|	color float float float
			{
				if(3 == ColorSamples)
					{
						RendermanInterface::RtColor Cq;
						Cq[0] = $2;
						Cq[1] = $3;
						Cq[2] = $4;
						ParseCallbackInterface->RiColor(Cq);
					}
				else
					{
						yyerror("Incorrect Number of Color Samples");
					}
			}
	|	opacity scalar_array
			{
				if($2->Count() == ColorSamples)
					ParseCallbackInterface->RiOpacity(&(*$2)[0]);
				else
					yyerror("Incorrect Number of Color Samples");
					
				DiscardArrayValue($2);
			}
	|	opacity float float float
			{
				if(3 == ColorSamples)
					{
						RendermanInterface::RtColor Cq;
						Cq[0]=$2;
						Cq[1]=$3;
						Cq[2]=$4;
						ParseCallbackInterface->RiOpacity(Cq);
					}
				else
					{
						yyerror("Incorrect Number of Color Samples");
					}
			}
	|	texturecoordinates float float float float float float float float
			{ ParseCallbackInterface->RiTextureCoordinates($2, $3, $4, $5, $6, $7, $8, $9); }
	|	texturecoordinates scalar_array
			{
				if($2->Count() == 8)
					ParseCallbackInterface->RiTextureCoordinates((*$2)[0], (*$2)[1], (*$2)[2], (*$2)[3], (*$2)[4], (*$2)[5], (*$2)[6], (*$2)[7]);
				else
					yyerror("Expecting 8 Values");
					
				DiscardArrayValue($2);
			}
	|	lightsource string integer opttvpairs
			{
				DeclareLightHandle($3, ParseCallbackInterface->RiLightSourceV($2, $4->Count(), $4->Tokens(), $4->Values()));
				DiscardStringValue($2);
				DiscardTokenValuePairs($4);
			}
	|	arealightsource string integer opttvpairs
			{
				DeclareLightHandle($3, ParseCallbackInterface->RiAreaLightSourceV($2, $4->Count(), $4->Tokens(), $4->Values()));
				DiscardStringValue($2);
				DiscardTokenValuePairs($4);
			}
	|	illuminate integer integer
			{
				RendermanInterface::RtLightHandle handle = LookupLightHandle($2);
				if(handle)
					ParseCallbackInterface->RiIlluminate(handle, $3 != 0);
				else
					yyerror("Undeclared Light");
			}
	|	surface string opttvpairs
			{
				ParseCallbackInterface->RiSurfaceV($2, $3->Count(), $3->Tokens(), $3->Values());
				DiscardStringValue($2);
				DiscardTokenValuePairs($3);
			}
	|	atmosphere string opttvpairs
			{
				ParseCallbackInterface->RiAtmosphereV($2, $3->Count(), $3->Tokens(), $3->Values());
				DiscardStringValue($2);
				DiscardTokenValuePairs($3);
			}
	|	interior string opttvpairs
			{
				ParseCallbackInterface->RiInteriorV($2, $3->Count(), $3->Tokens(), $3->Values());
				DiscardStringValue($2);
				DiscardTokenValuePairs($3);
			}
	|	exterior string opttvpairs
			{
				ParseCallbackInterface->RiExteriorV($2, $3->Count(), $3->Tokens(), $3->Values());
				DiscardStringValue($2);
				DiscardTokenValuePairs($3);
			}
	|	shadingrate float
			{ ParseCallbackInterface->RiShadingRate($2); }
	|	shadinginterpolation string
			{
				ParseCallbackInterface->RiShadingInterpolation((char*)($2));
				DiscardStringValue($2);
			}
	|	matte integer
			{ ParseCallbackInterface->RiMatte($2!=0); }
	|	bound	float float float float float float
			{
				RendermanInterface::RtBound b;
				b[0]=$2; b[1]=$3; b[2]=$4; b[3]=$5; b[4]=$6; b[5]=$7;
				ParseCallbackInterface->RiBound(b);
			}
	|	bound	scalar_array
			{
				if($2->Count() == 6)
					ParseCallbackInterface->RiBound(&(*$2)[0]);
				else
					yyerror("Expecting 6 Values");
					
				DiscardArrayValue($2);
			}
	|	detail float float float float float float
			{
				RendermanInterface::RtBound b;
				b[0]=$2; b[1]=$3; b[2]=$4; b[3]=$5; b[4]=$6; b[5]=$7;
				ParseCallbackInterface->RiDetail(b);
			}
	|	detail scalar_array
			{
				if($2->Count() == 6)
					ParseCallbackInterface->RiDetail(&(*$2)[0]);
				else
					yyerror("Expecting 6 Values");
				DiscardArrayValue($2);
			}
	|	detailrange float float float float
			{ ParseCallbackInterface->RiDetailRange($2, $3, $4, $5); }
	|	detailrange scalar_array
			{
				if($2->Count() == 4)
					ParseCallbackInterface->RiDetailRange((*$2)[0], (*$2)[1], (*$2)[2], (*$2)[3]);
				else
					yyerror("Expecting 4 Values");
				DiscardArrayValue($2);
			}
	|	geometricapproximation string float
			{
				ParseCallbackInterface->RiGeometricApproximation((char*)($2), $3);
				DiscardStringValue($2);
			}
	|	orientation string
			{
				ParseCallbackInterface->RiOrientation((char*)($2));
				DiscardStringValue($2);
			}
	|	reverseorientation
			{ ParseCallbackInterface->RiReverseOrientation(); }
	|	sides integer
			{ ParseCallbackInterface->RiSides($2); }
	|	identity
			{ ParseCallbackInterface->RiIdentity(); }
	|	transform scalar_array
			{
				RendermanInterface::RtMatrix Trans;
				if($2->Count() == 16)
					{
						Trans[0][0]=(*$2)[0];
						Trans[0][1]=(*$2)[1];
						Trans[0][2]=(*$2)[2];
						Trans[0][3]=(*$2)[3];
						Trans[1][0]=(*$2)[4];
						Trans[1][1]=(*$2)[5];
						Trans[1][2]=(*$2)[6];
						Trans[1][3]=(*$2)[7];
						Trans[2][0]=(*$2)[8];
						Trans[2][1]=(*$2)[9];
						Trans[2][2]=(*$2)[10];
						Trans[2][3]=(*$2)[11];
						Trans[3][0]=(*$2)[12];
						Trans[3][1]=(*$2)[13];
						Trans[3][2]=(*$2)[14];
						Trans[3][3]=(*$2)[15];
						ParseCallbackInterface->RiTransform(Trans);
					}
				else
					{
						yyerror("Expecting 16 Values");
					}
				DiscardArrayValue($2);
			}
	|	concattransform scalar_array
			{
				RendermanInterface::RtMatrix Trans;
				if($2->Count() == 16)
					{
						Trans[0][0]=(*$2)[0];
						Trans[0][1]=(*$2)[1];
						Trans[0][2]=(*$2)[2];
						Trans[0][3]=(*$2)[3];
						Trans[1][0]=(*$2)[4];
						Trans[1][1]=(*$2)[5];
						Trans[1][2]=(*$2)[6];
						Trans[1][3]=(*$2)[7];
						Trans[2][0]=(*$2)[8];
						Trans[2][1]=(*$2)[9];
						Trans[2][2]=(*$2)[10];
						Trans[2][3]=(*$2)[11];
						Trans[3][0]=(*$2)[12];
						Trans[3][1]=(*$2)[13];
						Trans[3][2]=(*$2)[14];
						Trans[3][3]=(*$2)[15];
						ParseCallbackInterface->RiConcatTransform(Trans);
					}
				else
					{
						yyerror("Expecting 16 Values");
					}
				DiscardArrayValue($2);
			}
	|	perspective float
			{ ParseCallbackInterface->RiPerspective($2); }
	|	translate	float float float
			{ ParseCallbackInterface->RiTranslate($2, $3, $4); }
	|	rotate float float float float
			{ ParseCallbackInterface->RiRotate($2, $3, $4, $5); }
	|	scale float float float
			{ ParseCallbackInterface->RiScale($2, $3, $4); }
	|	skew float float float float float float float
			{ ParseCallbackInterface->RiSkew($2, $3, $4, $5, $6, $7, $8); }
	|	skew scalar_array
			{
				if($2->Count( ) == 7)
					ParseCallbackInterface->RiSkew((*$2)[0], (*$2)[1], (*$2)[2], (*$2)[3], (*$2)[4], (*$2)[5], (*$2)[6]);
				else
					yyerror("Expecting 7 Values");
				DiscardArrayValue($2);
			}
	|	deformation string opttvpairs
			{
				ParseCallbackInterface->RiDeformationV($2, $3->Count(), $3->Tokens(), $3->Values());
				DiscardStringValue($2);
				DiscardTokenValuePairs($3);
			}
	|	displacement string opttvpairs
			{
				ParseCallbackInterface->RiDisplacementV($2, $3->Count(), $3->Tokens(), $3->Values());
				DiscardStringValue($2);
				DiscardTokenValuePairs($3);
			}
	|	coordinatesystem string
			{
				ParseCallbackInterface->RiCoordinateSystem($2);
				DiscardStringValue($2);
			}
	|	coordsystransform string
			{
				ParseCallbackInterface->RiCoordSysTransform($2);
				DiscardStringValue($2);
			}
	|	transformpoints
			{ printf("TRANSFORMPOINTS\n"); }
	|	transformbegin
			{ ParseCallbackInterface->RiTransformBegin(); }
	|	transformend
			{ ParseCallbackInterface->RiTransformEnd(); }
	|	attribute string opttvpairs
			{
				ParseCallbackInterface->RiAttributeV($2, $3->Count(), $3->Tokens(), $3->Values());
				DiscardStringValue($2);
				DiscardTokenValuePairs($3);
			}
	|	polygon tvpairs
			{
				int vertices = 0;
				for(int i = 0; i < $2->Count(); i++)
					{
						if($2->Tokens()[i] == std::string("P"))
							vertices = $2->Counts()[i] / 3;
					}
				ParseCallbackInterface->RiPolygonV(vertices, $2->Count(), $2->Tokens(), $2->Values());
				DiscardTokenValuePairs($2);
			}
	|	generalpolygon integer_array tvpairs
			{
				ParseCallbackInterface->RiGeneralPolygonV($2->Count(), &(*$2)[0], $3->Count(), $3->Tokens(), $3->Values());
				DiscardArrayValue($2);
				DiscardTokenValuePairs($3);
			}
	|	pointspolygons integer_array integer_array tvpairs
			{
				ParseCallbackInterface->RiPointsPolygonsV($2->Count(), &(*$2)[0], &(*$3)[0], $4->Count(), $4->Tokens(), $4->Values());
				DiscardArrayValue($2);
				DiscardArrayValue($3);
				DiscardTokenValuePairs($4);
			}
	|	pointsgeneralpolygons	integer_array integer_array integer_array tvpairs
			{
				ParseCallbackInterface->RiPointsGeneralPolygonsV($2->Count(), &(*$2)[0], &(*$3)[0], &(*$4)[0], $5->Count(), $5->Tokens(), $5->Values());
				DiscardArrayValue($2);
				DiscardArrayValue($3);
				DiscardArrayValue($4);
				DiscardTokenValuePairs($5);
			}
	|	basis string integer string integer
			{
/*
				RtBasis u,v;
				BasisFromName(u, $2);
				BasisFromName(v, $4);
				ParseCallbackInterface->RiBasis(u, $3,v, $5);
				DiscardStringValue($2);
				DiscardStringValue($4);
*/
			}
	|	basis string integer scalar_array integer
			{
/*
				if(CheckArrayType(Type_Float, $4) && $4.aftype->size()>=16)
					{
						RtBasis u,v;
						int i,j;
						for(i=0; i<4; i++)
							for(j=0; j<4; j++)
								v[i][j]=(*$4.aftype)[(i*4)+j];

						BasisFromName(u, $2);
						ParseCallbackInterface->RiBasis(u, $3,v, $5);
					}
				else
					{
						yyerror("Expecting 16 floats");
					}
				DiscardStringValue($2);
				DiscardArrayValue($4);
*/
									}
	|	basis scalar_array integer string integer
			{
/*
				if(CheckArrayType(Type_Float, $2) && $2.aftype->size()>=16)
					{

						RtBasis u,v;
						int i,j;
						for(i=0; i<4; i++)
							for(j=0; j<4; j++)
								u[i][j]=(*$2.aftype)[(i*4)+j];

						BasisFromName(v, $4);
						ParseCallbackInterface->RiBasis(u, $3,v, $5);
					}
				else
					{
						yyerror("Expecting 16 floats");
					}
				DiscardArrayValue($2);
				DiscardStringValue($4);
*/
									}
	|	basis scalar_array integer scalar_array integer
			{
				if($2->Count( ) == 16 && $4->Count() == 16)
					{
						// Build a matrix for the u and v directions
						RendermanInterface::RtBasis u, v;
						int i,j;
						for(i=0; i<4; i++)
							{
								for(j=0; j<4; j++)
									{
										u[i][j]=(*$2)[(i*4)+j];
										v[i][j]=(*$4)[(i*4)+j];
									}
							}
						ParseCallbackInterface->RiBasis(u, $3,v, $5);
					}
				else
					{
						yyerror("Expecting 16 Values");
					}
				DiscardArrayValue($2);
				DiscardArrayValue($4);
			}
	|	patch string tvpairs
			{
				ParseCallbackInterface->RiPatchV(const_cast<char*>($2), $3->Count(), $3->Tokens(), $3->Values());
				DiscardStringValue($2);
				DiscardTokenValuePairs($3);
			}
	|	patchmesh	string integer string integer string tvpairs
			{
				ParseCallbackInterface->RiPatchMeshV(const_cast<char*>($2), $3, const_cast<char*>($4), $5, const_cast<char*>($6), $7->Count(), $7->Tokens(), $7->Values());
				DiscardStringValue($2);
				DiscardStringValue($4);
				DiscardStringValue($6);
				DiscardTokenValuePairs($7);
			}
	|	nupatch integer integer scalar_array float float integer integer scalar_array float float tvpairs
			{
				ParseCallbackInterface->RiNuPatchV($2, $3, &(*$4)[0], $5, $6, $7, $8, &(*$9)[0], $10, $11, $12->Count(), $12->Tokens(), $12->Values());
				DiscardArrayValue($4);
				DiscardArrayValue($9);
				DiscardTokenValuePairs($12);
			}
	|	trimcurve integer_array integer_array scalar_array scalar_array scalar_array integer_array scalar_array scalar_array scalar_array
			{
				ParseCallbackInterface->RiTrimCurve($2->Count(), &(*$2)[0], &(*$3)[0], &(*$4)[0], &(*$5)[0], &(*$6)[0], &(*$7)[0], &(*$8)[0], &(*$9)[0], &(*$10)[0]);
				DiscardArrayValue($2);
				DiscardArrayValue($3);
				DiscardArrayValue($4);
				DiscardArrayValue($5);
				DiscardArrayValue($6);
				DiscardArrayValue($7);
				DiscardArrayValue($8);
				DiscardArrayValue($9);
				DiscardArrayValue($10);
			}
	|	sphere float float float float opttvpairs
			{
				ParseCallbackInterface->RiSphereV($2, $3, $4, $5, $6->Count(), $6->Tokens(), $6->Values());
				DiscardTokenValuePairs($6);
			}
	|	cone float float float opttvpairs
			{
				ParseCallbackInterface->RiConeV($2, $3, $4, $5->Count(), $5->Tokens(), $5->Values());
				DiscardTokenValuePairs($5);
			}
	|	cylinder float float float float opttvpairs
			{
				ParseCallbackInterface->RiCylinderV($2, $3, $4, $5, $6->Count(), $6->Tokens(), $6->Values());
				DiscardTokenValuePairs($6);
			}
	|	hyperboloid scalar_array opttvpairs
			{
				ParseCallbackInterface->RiHyperboloidV(&(*$2)[0], &(*$2)[3], (*$2)[6], $3->Count(), $3->Tokens(), $3->Values());
				DiscardArrayValue($2);
				DiscardTokenValuePairs($3);
			}
	|	hyperboloid float float float float float float float opttvpairs
			{
				RendermanInterface::RtPoint p1, p2;
				p1[0]=$2;
				p1[1]=$3;
				p1[2]=$4;
				p2[0]=$5;
				p2[1]=$6;
				p2[2]=$7;
				ParseCallbackInterface->RiHyperboloidV(p1, p2, $8, $9->Count(), $9->Tokens(), $9->Values());
				DiscardTokenValuePairs($9);
			}
	|	paraboloid float float float float opttvpairs
			{
				ParseCallbackInterface->RiParaboloidV($2, $3, $4, $5, $6->Count(), $6->Tokens(), $6->Values());
				DiscardTokenValuePairs($6);
			}
	|	disk float float float opttvpairs
			{
				ParseCallbackInterface->RiDiskV($2, $3, $4, $5->Count(), $5->Tokens(), $5->Values());
				DiscardTokenValuePairs($5);
			}
	|	torus float float float float float opttvpairs
			{
				ParseCallbackInterface->RiTorusV($2, $3, $4, $5, $6, $7->Count(), $7->Tokens(), $7->Values());
				DiscardTokenValuePairs($7);
			}
	|	geometry string opttvpairs
			{
				ParseCallbackInterface->RiGeometryV($2, $3->Count(), $3->Tokens(), $3->Values());
				DiscardStringValue($2);
				DiscardTokenValuePairs($3);
			}
	|	solidbegin string
			{
				ParseCallbackInterface->RiSolidBegin(const_cast<char*>($2));
				DiscardStringValue($2);
			}
	|	solidend
			{ ParseCallbackInterface->RiSolidEnd(); }
	|	objectbegin integer
			{
				// TODO: Record information about handle.
				ParseCallbackInterface->RiObjectBegin();
			}
	|	objectend
			{ ParseCallbackInterface->RiObjectEnd(); }
	|	objectinstance integer
			{
				// TODO: Find object pointer from handle.
				ParseCallbackInterface->RiObjectInstance(0);
			}
	|	motionbegin scalar_array
			{
				ParseCallbackInterface->RiMotionBeginV($2->Count(), &(*$2)[0]);
				DiscardArrayValue($2);
			}
	|	motionend
			{ ParseCallbackInterface->RiMotionEnd(); }
	|	maketexture	string string string string string float float opttvpairs
			{
				RendermanInterface::RtFilterFunc pFilter=RiGaussianFilter;
				// TODO: names shoudl be stored elsewhere.
				if(strcmp($6,"box")==0)		pFilter=RiBoxFilter;
				if(strcmp($6,"triangle")==0)	pFilter=RiTriangleFilter;
				if(strcmp($6,"sinc")==0)		pFilter=RiSincFilter;
				if(strcmp($6,"catmull-rom")==0) pFilter=RiCatmullRomFilter;
				if(strcmp($6,"disk")==0)		pFilter=RiDiskFilter;
				if(strcmp($6,"bessel")==0)		pFilter=RiBesselFilter;

				ParseCallbackInterface->RiMakeTextureV((char*)($2), (char*)($3), (char*)($4), (char*)($5), pFilter, $7, $8,
														$9->Count(), $9->Tokens(), $9->Values());
			}
	|	makebump string string string string string float float opttvpairs
			{
				RendermanInterface::RtFilterFunc pFilter=RiGaussianFilter;
				// TODO: names shoudl be stored elsewhere.
				if(strcmp($6,"box")==0)			pFilter=RiBoxFilter;
				if(strcmp($6,"triangle")==0)	pFilter=RiTriangleFilter;
				if(strcmp($6,"sinc")==0)		pFilter=RiSincFilter;
				if(strcmp($6,"catmull-rom")==0) pFilter=RiCatmullRomFilter;
				if(strcmp($6,"disk")==0) pFilter=RiDiskFilter;
				if(strcmp($6,"bessel")==0) pFilter=RiBesselFilter;

				ParseCallbackInterface->RiMakeBumpV((char*)($2), (char*)($3),(char*)($4), (char*)($5),pFilter,$7, $8,
													$9->Count(), $9->Tokens(), $9->Values());
			}
	|	makelatlongenvironment string string string float float opttvpairs
			{
				RendermanInterface::RtFilterFunc pFilter=RiGaussianFilter;
				// TODO: names shoudl be stored elsewhere.
				if(strcmp($4,"box")==0)		pFilter=RiBoxFilter;
				if(strcmp($4,"triangle")==0)	pFilter=RiTriangleFilter;
				if(strcmp($4,"sinc")==0)		pFilter=RiSincFilter;
				if(strcmp($4,"catmull-rom")==0) pFilter=RiCatmullRomFilter;
				if(strcmp($4,"disk")==0)		pFilter=RiDiskFilter;
				if(strcmp($4,"bessel")==0)		pFilter=RiBesselFilter;

				ParseCallbackInterface->RiMakeLatLongEnvironmentV((char*)($2), (char*)($3),pFilter,$5, $6,
																$7->Count(), $7->Tokens(), $7->Values());
			}
	|	makecubefaceenvironment	string string string string string string string float string float float opttvpairs
			{
				RendermanInterface::RtFilterFunc pFilter=RiGaussianFilter;
				// TODO: names should be stored elsewhere.
				if(strcmp($10,"box")==0)		pFilter=RiBoxFilter;
				if(strcmp($10,"triangle")==0)	pFilter=RiTriangleFilter;
				if(strcmp($10,"sinc")==0)		pFilter=RiSincFilter;
				if(strcmp($10,"catmull-rom")==0) pFilter=RiCatmullRomFilter;
				if(strcmp($10,"disk")==0)		pFilter=RiDiskFilter;
				if(strcmp($10,"bessel")==0)		pFilter=RiBesselFilter;

				ParseCallbackInterface->RiMakeCubeFaceEnvironmentV(
							 (char*)($2), (char*)($3),
							 (char*)($4), (char*)($5),
							 (char*)($6), (char*)($7),
							 (char*)($8),
							 $9,
							 pFilter,
							 $11, $12,
							 $13->Count(), $13->Tokens(), $13->Values());
			}
	|	makeshadow string string opttvpairs
			{
				ParseCallbackInterface->RiMakeShadowV($2, $3, $4->Count(), $4->Tokens(), $4->Values());
			}
	|	errorhandler string
			{
/*
				// TODO: names should be stored elsewhere.
				if(strcmp($2,"ignore")==0)
					ParseCallbackInterface->RiErrorHandler(&ParseCallbackInterface->RiErrorIgnore);
				else if(strcmp($2,"print")==0)
					ParseCallbackInterface->RiErrorHandler(&ParseCallbackInterface->RiErrorPrint);
				else if(strcmp($2,"abort")==0)
					ParseCallbackInterface->RiErrorHandler(&ParseCallbackInterface->RiErrorAbort);
*/
									}
	|	errorignore
			{ printf("ERRORIGNORE\n"); }
	|	errorprint
			{ printf("ERRORPRINT\n"); }
	|	errorabort
			{ printf("ERRORABORT\n"); }
	|	subdivisionmesh string integer_array integer_array string_array_or_empty integer_array_or_empty integer_array_or_empty scalar_array_or_empty tvpairs
			{
				char** ptags=new char*[$5->Count()];
				int i;
				for(i=0; i<$5->Count(); i++)
				{
					char* ptag=new char[strlen((*$5)[i])];
					strcpy(ptag, (*$5)[i]);
					ptags[i]=ptag;
				}

				ParseCallbackInterface->RiSubdivisionMeshV(const_cast<char*>($2),
								$3->Count(),&(*$3)[0], &(*$4)[0],
								$5->Count(), ptags,
								&(*$6)[0],
								&(*$7)[0],
								&(*$8)[0],
								$9->Count(), $9->Tokens(), $9->Values());

				for(i=0; i<$5->Count(); i++)
					delete[](ptags[i]);
				delete[](ptags);

				DiscardStringValue($2);
				DiscardArrayValue($3);
				DiscardArrayValue($4);
				DiscardArrayValue($5);
				DiscardArrayValue($6);
				DiscardArrayValue($7);
				DiscardArrayValue($8);
				DiscardTokenValuePairs($9);
			}
	|	subdivisionmesh string integer_array integer_array tvpairs
			{
				ParseCallbackInterface->RiSubdivisionMeshV(const_cast<char*>($2), $3->Count(), &(*$3)[0], &(*$4)[0], 0, NULL, NULL, NULL, NULL, $5->Count(), $5->Tokens(), $5->Values());
				DiscardStringValue($2);
				DiscardArrayValue($3);
				DiscardArrayValue($4);
				DiscardTokenValuePairs($5);
			}
	|	UNKNOWN_TOKEN
			{
				// Print the error, then force the scanner into 'request'
				// mode which will make it ignore all parameters until the
				// next request token.
				yyerror("Unrecognised RIB request");
				ExpectRequest();
			}
	;

arealightsource : REQUEST_TOKEN_AREALIGHTSOURCE	{ ExpectParams(); }
atmosphere : REQUEST_TOKEN_ATMOSPHERE	{ ExpectParams(); }
attribute : REQUEST_TOKEN_ATTRIBUTE	{ ExpectParams(); }
attributebegin : REQUEST_TOKEN_ATTRIBUTEBEGIN	{ ExpectParams(); }
attributeend : REQUEST_TOKEN_ATTRIBUTEEND	{ ExpectParams(); }
basis : REQUEST_TOKEN_BASIS	{ ExpectParams(); }
bound : REQUEST_TOKEN_BOUND	{ ExpectParams(); }
clipping : REQUEST_TOKEN_CLIPPING	{ ExpectParams(); }
color : REQUEST_TOKEN_COLOR	{ ExpectParams(); }
colorsamples : REQUEST_TOKEN_COLORSAMPLES	{ ExpectParams(); }
concattransform : REQUEST_TOKEN_CONCATTRANSFORM	{ ExpectParams(); }
cone : REQUEST_TOKEN_CONE	{ ExpectParams(); }
coordinatesystem : REQUEST_TOKEN_COORDINATESYSTEM	{ ExpectParams(); }
coordsystransform : REQUEST_TOKEN_COORDSYSTRANSFORM	{ ExpectParams(); }
cropwindow : REQUEST_TOKEN_CROPWINDOW	{ ExpectParams(); }
cylinder : REQUEST_TOKEN_CYLINDER	{ ExpectParams(); }
declare : REQUEST_TOKEN_DECLARE	{ ExpectParams(); }
deformation : REQUEST_TOKEN_DEFORMATION	{ ExpectParams(); }
depthoffield : REQUEST_TOKEN_DEPTHOFFIELD	{ ExpectParams(); }
detail : REQUEST_TOKEN_DETAIL	{ ExpectParams(); }
detailrange : REQUEST_TOKEN_DETAILRANGE	{ ExpectParams(); }
disk : REQUEST_TOKEN_DISK	{ ExpectParams(); }
displacement : REQUEST_TOKEN_DISPLACEMENT	{ ExpectParams(); }
display : REQUEST_TOKEN_DISPLAY	{ ExpectParams(); }
errorabort : REQUEST_TOKEN_ERRORABORT	{ ExpectParams(); }
errorhandler : REQUEST_TOKEN_ERRORHANDLER	{ ExpectParams(); }
errorignore : REQUEST_TOKEN_ERRORIGNORE	{ ExpectParams(); }
errorprint : REQUEST_TOKEN_ERRORPRINT	{ ExpectParams(); }
exposure : REQUEST_TOKEN_EXPOSURE	{ ExpectParams(); }
exterior : REQUEST_TOKEN_EXTERIOR	{ ExpectParams(); }
format : REQUEST_TOKEN_FORMAT	{ ExpectParams(); }
frameaspectratio : REQUEST_TOKEN_FRAMEASPECTRATIO	{ ExpectParams(); }
framebegin : REQUEST_TOKEN_FRAMEBEGIN	{ ExpectParams(); }
frameend : REQUEST_TOKEN_FRAMEEND	{ ExpectParams(); }
generalpolygon : REQUEST_TOKEN_GENERALPOLYGON	{ ExpectParams(); }
geometricapproximation : REQUEST_TOKEN_GEOMETRICAPPROXIMATION	{ ExpectParams(); }
geometry : REQUEST_TOKEN_GEOMETRY	{ ExpectParams(); }
hider : REQUEST_TOKEN_HIDER	{ ExpectParams(); }
hyperboloid : REQUEST_TOKEN_HYPERBOLOID	{ ExpectParams(); }
identity : REQUEST_TOKEN_IDENTITY	{ ExpectParams(); }
illuminate : REQUEST_TOKEN_ILLUMINATE	{ ExpectParams(); }
imager : REQUEST_TOKEN_IMAGER	{ ExpectParams(); }
interior : REQUEST_TOKEN_INTERIOR	{ ExpectParams(); }
lightsource : REQUEST_TOKEN_LIGHTSOURCE	{ ExpectParams(); }
makebump : REQUEST_TOKEN_MAKEBUMP	{ ExpectParams(); }
makecubefaceenvironment : REQUEST_TOKEN_MAKECUBEFACEENVIRONMENT	{ ExpectParams(); }
makelatlongenvironment : REQUEST_TOKEN_MAKELATLONGENVIRONMENT	{ ExpectParams(); }
makeshadow : REQUEST_TOKEN_MAKESHADOW	{ ExpectParams(); }
maketexture : REQUEST_TOKEN_MAKETEXTURE	{ ExpectParams(); }
matte : REQUEST_TOKEN_MATTE	{ ExpectParams(); }
motionbegin : REQUEST_TOKEN_MOTIONBEGIN	{ ExpectParams(); }
motionend : REQUEST_TOKEN_MOTIONEND	{ ExpectParams(); }
nupatch : REQUEST_TOKEN_NUPATCH	{ ExpectParams(); }
objectbegin : REQUEST_TOKEN_OBJECTBEGIN	{ ExpectParams(); }
objectend : REQUEST_TOKEN_OBJECTEND	{ ExpectParams(); }
objectinstance : REQUEST_TOKEN_OBJECTINSTANCE	{ ExpectParams(); }
opacity : REQUEST_TOKEN_OPACITY	{ ExpectParams(); }
option : REQUEST_TOKEN_OPTION	{ ExpectParams(); }
orientation : REQUEST_TOKEN_ORIENTATION	{ ExpectParams(); }
paraboloid : REQUEST_TOKEN_PARABOLOID	{ ExpectParams(); }
patch : REQUEST_TOKEN_PATCH	{ ExpectParams(); }
patchmesh : REQUEST_TOKEN_PATCHMESH	{ ExpectParams(); }
perspective : REQUEST_TOKEN_PERSPECTIVE	{ ExpectParams(); }
pixelfilter : REQUEST_TOKEN_PIXELFILTER	{ ExpectParams(); }
pixelsamples : REQUEST_TOKEN_PIXELSAMPLES	{ ExpectParams(); }
pixelvariance : REQUEST_TOKEN_PIXELVARIANCE	{ ExpectParams(); }
pointsgeneralpolygons : REQUEST_TOKEN_POINTSGENERALPOLYGONS	{ ExpectParams(); }
pointspolygons : REQUEST_TOKEN_POINTSPOLYGONS	{ ExpectParams(); }
polygon : REQUEST_TOKEN_POLYGON	{ ExpectParams(); }
projection : REQUEST_TOKEN_PROJECTION	{ ExpectParams(); }
quantize : REQUEST_TOKEN_QUANTIZE	{ ExpectParams(); }
relativedetail : REQUEST_TOKEN_RELATIVEDETAIL	{ ExpectParams(); }
reverseorientation : REQUEST_TOKEN_REVERSEORIENTATION	{ ExpectParams(); }
rotate : REQUEST_TOKEN_ROTATE	{ ExpectParams(); }
scale : REQUEST_TOKEN_SCALE	{ ExpectParams(); }
screenwindow : REQUEST_TOKEN_SCREENWINDOW	{ ExpectParams(); }
shadinginterpolation : REQUEST_TOKEN_SHADINGINTERPOLATION	{ ExpectParams(); }
shadingrate : REQUEST_TOKEN_SHADINGRATE	{ ExpectParams(); }
shutter : REQUEST_TOKEN_SHUTTER	{ ExpectParams(); }
sides : REQUEST_TOKEN_SIDES	{ ExpectParams(); }
skew : REQUEST_TOKEN_SKEW	{ ExpectParams(); }
solidbegin : REQUEST_TOKEN_SOLIDBEGIN	{ ExpectParams(); }
solidend : REQUEST_TOKEN_SOLIDEND	{ ExpectParams(); }
sphere : REQUEST_TOKEN_SPHERE	{ ExpectParams(); }
subdivisionmesh : REQUEST_TOKEN_SUBDIVISIONMESH	{ ExpectParams(); }
surface : REQUEST_TOKEN_SURFACE	{ ExpectParams(); }
texturecoordinates : REQUEST_TOKEN_TEXTURECOORDINATES	{ ExpectParams(); }
torus : REQUEST_TOKEN_TORUS	{ ExpectParams(); }
transform : REQUEST_TOKEN_TRANSFORM	{ ExpectParams(); }
transformbegin : REQUEST_TOKEN_TRANSFORMBEGIN	{ ExpectParams(); }
transformend : REQUEST_TOKEN_TRANSFORMEND	{ ExpectParams(); }
transformpoints : REQUEST_TOKEN_TRANSFORMPOINTS	{ ExpectParams(); }
translate : REQUEST_TOKEN_TRANSLATE	{ ExpectParams(); }
trimcurve : REQUEST_TOKEN_TRIMCURVE	{ ExpectParams(); }
version : REQUEST_TOKEN_VERSION	{ ExpectParams(); }
worldbegin : REQUEST_TOKEN_WORLDBEGIN	{ ExpectParams(); }
worldend : REQUEST_TOKEN_WORLDEND	{ ExpectParams(); }


float : FLOAT_TOKEN
	|	integer
			{ $$ = static_cast<float>($1); }
	;

floats :	FLOAT_TOKEN
			{
				$$ = new FloatArray($1);
			}
	|	floats FLOAT_TOKEN
			{
				$$ = $1;
				$$->push_back($2);
			}
	|	floats INTEGER_TOKEN
			{
				$$ = $1;
				$$->push_back($2);
			}
	|	integers FLOAT_TOKEN
			{
				$$ = new FloatArray($1);
				$$->push_back($2);
				DiscardArrayValue($1);
			}
	;


float_array : '[' floats ']'
			{
				$$ = $2;
			}
	;

integer : INTEGER_TOKEN
			{
				$$ = $1;
			}
	;

integers : INTEGER_TOKEN
			{
				$$ = new IntegerArray($1);
			}
	|	integers integer
			{
				$$ = $1;
				$$->push_back($2);
			}

integer_array : '[' integers ']'
			{
				$$ = $2;
			}
	;

integer_array_or_empty : integer_array
	|	'[' ']'
			{
				$$ = new IntegerArray;
			}
	;

string : STRING_TOKEN
			{
				$$ = $1;
			}
	;

strings : STRING_TOKEN
			{
				$$ = new StringArray($1);
			}
	|	strings STRING_TOKEN
			{
				$$ = $1;
				$$->push_back($2);
			}

string_array : '[' strings ']'
			{
				$$ = $2;
			}
	;

string_array_or_empty : string_array
	|	'[' ']'
			{
				$$ = new StringArray;
			}
	;

scalar_array : float_array
	|	integer_array
			{
				$$ = new FloatArray($1);
				DiscardArrayValue($1);
			}
	;


scalar_array_or_empty : scalar_array
	|	'[' ']'
			{
				$$ = new FloatArray;
			}
	;

tvpair	: string INTEGER_TOKEN
			{
				// Set default value ...
				$$ = 0;

				switch(LookupParameterType($1))
					{
						case Type_Unknown:
							yyerror("Undeclared Integer Parameter [" + std::string($1) + "]");
							DiscardStringValue($1);
							break;

						case Type_Integer:
							// We don't have to discard $1 because the TokenValuePair assumes ownership of the string ...
							$$ = new TokenValuePair($1, new IntegerArray($2));
							break;

						// This is where we perform type conversion from integer to float, based on the declared parameter type ...
						case Type_Float:
						case Type_Point:
						case Type_Color:
						case Type_hPoint:
						case Type_Normal:
						case Type_Vector:
						case Type_Matrix:
							// We don't have to discard $1 because the TokenValuePair assumes ownership of the string ...
							$$ = new TokenValuePair($1, new FloatArray(static_cast<RendermanInterface::RtFloat>($2)));
							break;

						default:
							yyerror("Integer Parameter Type Mismatch [" + std::string($1) + "]");
							DiscardStringValue($1);
							break;
					}
			}
	|	string integer_array
			{
				// Set default value ...
				$$ = 0;

				switch(LookupParameterType($1))
					{
						case Type_Unknown:
							yyerror("Undeclared Integer Array Parameter [" + std::string($1) + "]");
							DiscardStringValue($1);
							DiscardArrayValue($2);
							break;

						case Type_Integer:
							// We don't have to discard $1 because the TokenValuePair assumes ownership of the string ...
							$$ = new TokenValuePair($1, $2);
							break;

						// This is where we perform type conversion from integer to float, based on the declared parameter type ...
						case Type_Float:
						case Type_Point:
						case Type_Color:
						case Type_hPoint:
						case Type_Normal:
						case Type_Vector:
						case Type_Matrix:
							// We don't have to discard $1 because the TokenValuePair assumes ownership of the string ...
							$$ = new TokenValuePair($1, new FloatArray($2));
							DiscardArrayValue($2);
							break;

						default:
							yyerror("Integer Array Parameter Type Mismatch [" + std::string($1) + "]");
							DiscardStringValue($1);
							DiscardArrayValue($2);
							break;
					}
			}
	|	string FLOAT_TOKEN
			{
				// Set default value ...
				$$ = 0;

				switch(LookupParameterType($1))
					{
						case Type_Unknown:
							yyerror("Undeclared Float Parameter: [" + std::string($1) + "]");
							DiscardStringValue($1);
							break;

						case Type_Float:
							// We don't have to discard $1 because the TokenValuePair assumes ownership of the string ...
							$$ = new TokenValuePair($1, new FloatArray($2));
							break;

						default:
							yyerror("Float Parameter Type Mismatch [" + std::string($1) + "]");
							DiscardStringValue($1);
							break;
					}
			}
	|	string float_array
			{
				// Set default value ...
				$$ = 0;

				switch(LookupParameterType($1))
					{
						case Type_Unknown:
							yyerror("Undeclared Float Array Parameter: [" + std::string($1) + "]");
							DiscardStringValue($1);
							DiscardArrayValue($2);
							break;

						case Type_Float:
						case Type_Point:
						case Type_Color:
						case Type_hPoint:
						case Type_Normal:
						case Type_Vector:
						case Type_Matrix:
							// We don't have to discard $1 because the TokenValuePair assumes ownership of the string ...
							$$ = new TokenValuePair($1, $2);
							break;

						default:
							yyerror("Float Array Parameter Type Mismatch [" + std::string($1) + "]");
							DiscardStringValue($1);
							DiscardArrayValue($2);
							break;
					}
			}
	|	string STRING_TOKEN
			{
				// Set default value ...
				$$ = 0;

				switch(LookupParameterType($1))
					{
						case Type_Unknown:
							yyerror("Undeclared String Parameter [" + std::string($1) + "]");
							DiscardStringValue($1);
							DiscardStringValue($2);
							break;

						case Type_String:
							// We don't have to discard $1 because the TokenValuePair assumes ownership of the string ...
							$$ = new TokenValuePair($1, new StringArray($2));
							break;

						default:
							yyerror("String Parameter Type Mismatch [" + std::string($1) + "]");
							DiscardStringValue($1);
							DiscardStringValue($2);
							break;
					}
			}
	|	string string_array
			{
				// Set default value ...
				$$ = 0;

				switch(LookupParameterType($1))
					{
						case Type_Unknown:
							yyerror("Undeclared String Array Parameter [" + std::string($1) + "]");
							DiscardStringValue($1);
							DiscardArrayValue($2);
							break;

						case Type_String:
							// We don't have to discard $1 because the TokenValuePair assumes ownership of the string ...
							$$ = new TokenValuePair($1, $2);
							break;

						default:
							yyerror("String Array Parameter Type Mismatch [" + std::string($1) + "]");
							DiscardStringValue($1);
							DiscardArrayValue($2);
							break;
					}
			}
	;

opttvpairs : /* Empty */
			{
				$$ = new TokenValuePairs;
			}
	|	error
			{
				$$ = new TokenValuePairs;
			}
	|	tvpairs
	;

tvpairs
	:	tvpair
			{
				// We don't need to delete $1 here, because the new TokenValuePairs object owns it, now ...
				$$ = new TokenValuePairs($1);
			}
	|	error tvpair
			{
				// We don't need to delete $2 here, because the new TokenValuePairs object owns it, now ...
				$$ = new TokenValuePairs($2);
			}
	|	tvpairs tvpair
			{
				// We don't need to delete $1 here, because we transfer its ownership to $$ ...
				// We don't need to delete $2 here, because the TokenValuePairs object in $$ owns it, now ...
				$$ = $1;
				$$->AddPair($2);
			}
	|	tvpairs error
	;

%%

static void yyerror(const std::string Message)
{
	ParseSucceeded = false;
	(*ParseErrorStream) << "librib > parser > error: " << Message << " at " << ParseStreamName << " line " << ParseLineNumber << std::endl;
}

static void ExpectRequest()
{
	fRequest=true;
}

static void ExpectParams()
{
	fParams=true;
}

static void DiscardStringValue(char* const Value)
{
	delete[] Value;
}

static void DiscardTokenValuePairs(TokenValuePairs* const Value)
{
	delete Value;
}

static void DiscardArrayValue(Array* const Value)
{
	delete Value;
}

static bool notspace(char C)
{
	return !isspace(C);
}

#ifdef AQSIS_SYSTEM_MACOSX
static bool isspacef(char C)
{
    return isspace(C);
}
#endif

typedef std::vector<std::string> Strings;
typedef Strings::iterator StringIterator;
static Strings Words(const std::string& Source)
{
	Strings results;

	std::string::const_iterator i = Source.begin();
	while(i != Source.end())
		{
			i = std::find_if(i, Source.end(), notspace);
			if(i == Source.end())
				break;
			
#ifdef AQSIS_SYSTEM_MACOSX
                        std::string::const_iterator j = std::find_if(i, Source.end(), isspacef);
#else
                        std::string::const_iterator j = std::find_if(i, Source.end(), isspace);
#endif

			
			results.push_back(std::string(i, j));
			
			i = j;
		}
	
	return results;
}

static std::string CleanParameterType(const std::string& Source)
{
	return Source.substr(0, Source.find("["));
}

static ParameterType ParseParameterType(const std::string& Source)
{
	if(0 == Source.size())
		{
			yyerror("Cannot Parse Empty Parameter Type");
			return Type_Unknown;
		}

	// Break the source string into whitespace delimited "words" ...
	Strings words(Words(Source));

	// Figure out the parameter type ...
	for(StringIterator word = words.begin(); word != words.end(); word++)
		{
			std::string cleanword = CleanParameterType(*word);
		
			if(cleanword == "integer")
				return Type_Integer;

			if(cleanword == "float")
				return Type_Float;
				
			if(cleanword == "hpoint")
				return Type_hPoint;
			
			if(cleanword == "point")
				return Type_Point;
				
			if(cleanword == "color")
				return Type_Color;
				
			if(cleanword == "vector")
				return Type_Vector;
				
			if(cleanword == "normal")
				return Type_Normal;
				
			if(cleanword == "matrix")
				return Type_Matrix;
				
			if(cleanword == "string")
				return Type_String;
		}

	return Type_Unknown;
}

typedef std::map<RendermanInterface::RtInt, RendermanInterface::RtLightHandle> LightHandleMap;
typedef LightHandleMap::iterator LightHandleIterator;
static LightHandleMap LightMap;

static void DeclareLightHandle(const RendermanInterface::RtInt ID, const RendermanInterface::RtLightHandle Handle)
{
	if(0 == Handle)
		{
			yyerror("NULL Light Handle");
			return;
		}

	if(LightMap.find(ID) != LightMap.end())
		{
			yyerror("Duplicate Light ID Will Be Ignored");
			return ;
		}
		
	LightMap[ID] = Handle;
}

static RendermanInterface::RtLightHandle LookupLightHandle(const RendermanInterface::RtInt ID)
{
	if(LightMap.find(ID) == LightMap.end())
		{
			yyerror("Undeclared Light ID");
			return 0;
		}
		
	return LightMap[ID];
}

typedef std::map<std::string, ParameterType> ParameterTypeMap;
typedef ParameterTypeMap::iterator ParameterTypeIterator;
static ParameterTypeMap TypeMap;

static ParameterType DeclareParameterType(const std::string& Name, const std::string& Type)
{
	if(0 == Name.size())
		{
			yyerror("Cannot Declare Unnamed Type");
			return Type_Unknown;
		}

	const ParameterType type = ParseParameterType(Type);

	if(type != Type_Unknown)
		TypeMap[Name] = type;

	return type;
}

static ParameterType LookupParameterType(const std::string& Name)
{
	if(0 == Name.size())
		{
			yyerror("Cannot Lookup Unnamed Type");
			return Type_Unknown;
		}

	// See if this parameter has already been declared ...
	ParameterTypeIterator parameter = TypeMap.find(Name);
	if(parameter != TypeMap.end())
		return static_cast<ParameterType>(parameter->second & Type_Mask);

	// Check to see if the type declaration is part of the parameter name ...
	return ParseParameterType(Name);
}

namespace librib
{

void ParserDeclare(librib::RendermanInterface& CallbackInterface, const std::string Name, const std::string Type)
{
	if(Type_Unknown == DeclareParameterType(Name, Type))
		yyerror("RiDeclare: Unknown type [" + Name + ", " + Type + "]");
		
	CallbackInterface.RiDeclare(Name.c_str(), Type.c_str());
}

}; // namespace librib


